package au.com.forward.SDtoCSV;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Date;

import au.com.forward.logging.*;

public class Main {

	static final int highestSupportedMetaVersion = 0;
	static final String MET_EXT = ".met";
	static final String BIN_EXT = ".bin";
	static final String CSV_EXT = ".csv";
	static final int NAME_DIM = 16;

	static final PrintStream oldStdOut = System.out;
	static boolean littleEndian = true;
	static int versionNo = -1; // not set
	static long sampleIdx = 0; // sample counter

	static long byteOrder; // 0x0000FFFF Note NOT Unicode marker which is
							// 0000fffe
	// this is byteOrder for version 0
	// for version 1 it is 0x0001FFFF, for version 2 it is 0x0002FFFF etc
	static long sampleInterval; // Sample interval in uS
	static long adcBits; // Size of ADC values.
	static long adcAvgs; // Number of ADC readings averaged 0, 4, 8, 16 or 32
	static long pinCount; // Number of analog pins in a sample.
	static long samplesPerBlock; // = SAMPLES_PER_BLOCK; // i.e. save PIN_COUNT
									// *
	// SAMPLES_PER_BLOCK values in a block == block
	// count
	static long noBufferBlocks; // Number of buffer blocks
	static long duration; // length of log data run, mS
	static long runTime; // how long to run this logging
	static long maxBuffersUsed; // max number of buffers saved at one time by
	// storeSampleBuffers()
	static long maxLatency; // max time spent in storeSampleBuffers(). Includes
	// closing/opening files when swapping files uS
	static long maxFileOpenTime; // max time spent closing/opening files when
									// swapping
	// files uS
	static long missedTimers;
	static long missedSamplesTotal;
	static long totalBlocksWritten;
	static long totalSamplesWritten; // divide by pinCount to output lines
	String currentFile; // [NAME_DIM]; // +1 FOR NULL
	static long pinNumbers[]; // 40 List of pin numbers in a sample.

	static {
		// set up logging also redirects System.err and System.out
		// throws a LoggingException if problem with setting up Java logging
		// put all log files in user's home dir
		// roll log to .old when it exceed 100K
		String userHome = "." + System.getProperty("file.separator", "/");
		Logging.initializeLogging(userHome + "SDtoCSV.log",
				"  ===  Log File for SD_Logging to CSV conversion "
						+ new Date(), false, 100);
		// do not redirect System.out as well as System.err and append to
		// existing log
	}

	public static void main(String[] args) {
		try {
			System.out
					.println("SD_Logging to CSV conversion.  Check SDtoCSV.log for progress and errors messages.");
			if (args.length < 1) {
				System.out.println(" Error!! missing <basefilename> argument.");
				System.out
						.println(" Usage: java -jar SDtoCSV.jar <basefilename>");
				System.out.flush();
			}

			LogStdStreams.getLogStream().println(
					" To see progress messages on Console use");
			LogStdStreams.getLogStream().println(
					"    java -jar SDtoCSV.jar <basefilename>");
			LogStdStreams.getLogStream().flush();

			if (args.length < 1) {
				System.err.println(" Error! missing <basefilename> argument.");
				System.err
						.println(" Usage: java -jar SDtoCSV.jar <basefilename>");
				System.exit(1);
			}
			// strip extension
			String baseFileName = args[0];
			int idx = baseFileName.indexOf('.');
			if (idx > -1) {
				// strip
				baseFileName = baseFileName.substring(0, idx).trim();
			}
			// System.err.println(" trimmed  :"+baseFileName);
			if (baseFileName.length() == 0) {
				System.err
						.println(" Base File Name is empty after removing extension");
				System.err.println(" Usage: ToCSV <basefilename>");
				System.exit(1);
			}
			LogStdStreams.getLogStream().println(
					" Base File Name '" + baseFileName + "'");
			LogStdStreams.getLogStream().flush();

			// System.out.print(baseFileName);
			// open and read the .met file
			File metFile = new File(baseFileName + MET_EXT);
			// System.out.print(metFile);
			final int BUFFER_SIZE = 512;
			ByteBuffer metaBuffer = ByteBuffer.allocate(BUFFER_SIZE);

			FileInputStream metaStream = null;
			try {
				metaStream = new FileInputStream(metFile);
			} catch (FileNotFoundException e) {
				System.err.println(" Error! Could not find metadata file");
				System.err.println(" " + metFile.getAbsolutePath());
				System.exit(1);
			}
			BufferedInputStream bufferedStream = null;
			bufferedStream = new BufferedInputStream(metaStream);
			int bytesRead = 0;
			try {
				bytesRead = bufferedStream.read(metaBuffer.array(), 0,
						BUFFER_SIZE);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				System.err.println(" Error! Reading "
						+ metFile.getAbsolutePath());
				System.err.println(" " + e.getMessage());
				System.exit(1);
			}
			if (bufferedStream != null) {
				try {
					bufferedStream.close();
				} catch (IOException e) {
					System.err.println(" Warning - error closing  "
							+ metFile.getAbsolutePath());
					System.err.println(" " + e.getMessage());
				}
			}
			if (metaStream != null) {
				try {
					metaStream.close();
				} catch (IOException e) {
					System.err.println(" Warning - error closing  "
							+ metFile.getAbsolutePath());
					System.err.println(" " + e.getMessage());
				}
			}

			// System.out.print(" read bytes:"+bytesRead); 240
			// read metaFile into dataBuffer.open()
			// look at endian
			if (bytesRead != 240) {
				System.err.println(" Error! Reading "
						+ metFile.getAbsolutePath());
				System.err.println(" incorrect file size");
				System.exit(1);
			}
			// metaBuffer.position(0);
			// for (int i = 0; i < 4; i++) {
			// short b = (short) metaBuffer.get(i);
			// b = (short) (b & 0x0000ff);
			// System.out.println(Integer.toHexString(b));
			// }
			littleEndian = false;
			metaBuffer.position(0);
			metaBuffer.order(ByteOrder.BIG_ENDIAN);
			int bo = metaBuffer.getInt(0);
			if ((bo & 0xffff0000) == 0xffff0000) {
				// System.out.println("i & 0xffff0000 " + (i & 0xffff0000));
				littleEndian = true;
			} else if ((bo & 0x0000ffff) == 0x0000ffff) {
				// System.out.println("(i & 0x0000ffff) " + (i & 0x0000ffff));
				littleEndian = false;
			} else {
				System.err.println(" Error! Reading "
						+ metFile.getAbsolutePath());
				System.err.println(" incorrect byte order entry");
				System.exit(1);

			}
			// System.out.println("Metadata file is in "
			// + (littleEndian ? "little endian" : "big endian"));

			// check for ffffff and version no
			if (!littleEndian) {
				// big endian
				metaBuffer.position(0);
				metaBuffer.order(ByteOrder.BIG_ENDIAN);
				short version = metaBuffer.getShort();
				// s1 should be 0xffff == -1
				versionNo = version;
				if (versionNo != 0) {
					System.err.println(" Error! Reading "
							+ metFile.getAbsolutePath());
					System.err.println(" unsupported version : " + versionNo);
					System.exit(1);
				}
			} else {
				// littleEndian
				metaBuffer.position(0);
				metaBuffer.order(ByteOrder.LITTLE_ENDIAN);
				metaBuffer.getShort();
				short version = metaBuffer.getShort();
				// s0 should be 0xffff == -1
				// need to swap
				versionNo = version;
				if (versionNo != 0) {
					System.err.println(" Error! Reading "
							+ metFile.getAbsolutePath());
					System.err.println(" unsupported version : " + versionNo);
					System.exit(1);
				}
			}
			// read in meta data
			// reset postion and endian
			metaBuffer.position(0);
			metaBuffer.order(littleEndian ? ByteOrder.LITTLE_ENDIAN
					: ByteOrder.BIG_ENDIAN);

			// const size_t NAME_DIM = 13;
			// // First block of file.
			// struct metadata_t {
			// uint32_t byteOrder; // 0x0000FFFF Note NOT Unicode marker which
			// is
			// 0000fffe
			// // Note: this is version 0, for version 1 use 0x0001FFFF, for
			// version
			// 2 use 0x0002FFFF etc
			// uint32_t sampleInterval; // Sample interval in uS
			// uint32_t adcBits; // Size of ADC values.
			// uint32_t adcAvgs; // Number of ADC readings averaged 0, 4, 8, 16
			// or
			// 32
			// uint32_t pinCount; // Number of analog pins in a sample.
			// uint32_t samplesPerBlock; // = SAMPLES_PER_BLOCK; // i.e. save
			// PIN_COUNT * SAMPLES_PER_BLOCK values in a block == block count
			// uint32_t noBufferBlocks; // Number of buffer blocks
			// uint32_t duration; // length of log data run, mS
			// uint32_t runTime; // how long to run this logging
			// uint32_t maxBuffersUsed; // max number of buffers saved at one
			// time
			// by storeSampleBuffers()
			// uint32_t maxLatency; // max time spent in storeSampleBuffers().
			// Includes closing/opening files when swapping files uS
			// uint32_t maxFileOpenTime; // max time spent closing/opening files
			// when swapping files uS
			// uint32_t missedTimers;
			// uint32_t missedSamplesTotal;
			// uint32_t totalBlocksWritten;
			// uint32_t totalSamplesWritten;
			// char currentFile[NAME_DIM]; // +1 FOR NULL
			// uint32_t pinNumber[40]; // List of pin numbers in a sample.

			byteOrder = uint32_tToLong(metaBuffer.getInt()); // 0x0000FFFF Note
																// NOT
																// Unicode
																// marker
																// which is
																// 0000fffe
			sampleInterval = uint32_tToLong(metaBuffer.getInt()); // Sample
																	// interval
																	// in uS
			adcBits = uint32_tToLong(metaBuffer.getInt()); // Size of ADC
															// values.
			adcAvgs = uint32_tToLong(metaBuffer.getInt()); // Number of ADC
															// readings
															// averaged 0, 4, 8,
															// 16
															// or 32
			pinCount = uint32_tToLong(metaBuffer.getInt()); // Number of analog
															// pins
															// in a sample.
			samplesPerBlock = uint32_tToLong(metaBuffer.getInt()); // =
																	// SAMPLES_PER_BLOCK;
																	// // i.e.
																	// save
																	// PIN_COUNT
																	// *
			// SAMPLES_PER_BLOCK values in a block == block
			// count
			noBufferBlocks = uint32_tToLong(metaBuffer.getInt()); // Number of
																	// buffer
																	// blocks
			duration = uint32_tToLong(metaBuffer.getInt()); // length of log
															// data
															// run, mS
			runTime = uint32_tToLong(metaBuffer.getInt()); // how long to run
															// this
															// logging
			maxBuffersUsed = uint32_tToLong(metaBuffer.getInt()); // max number
																	// of
																	// buffers
																	// saved
																	// at one
																	// time
																	// by
			// storeSampleBuffers()
			maxLatency = uint32_tToLong(metaBuffer.getInt()); // max time spent
																// in
																// storeSampleBuffers().
																// Includes
			// closing/opening files when swapping files uS
			maxFileOpenTime = uint32_tToLong(metaBuffer.getInt()); // max time
																	// spent
																	// closing/opening
																	// files
																	// when
																	// swapping
			// files uS
			missedTimers = uint32_tToLong(metaBuffer.getInt());
			missedSamplesTotal = uint32_tToLong(metaBuffer.getInt());
			totalBlocksWritten = uint32_tToLong(metaBuffer.getInt());
			totalSamplesWritten = uint32_tToLong(metaBuffer.getInt());
			// skip the next NAME_DIM bytes
			for (int k = 0; k < NAME_DIM; k++) {
				metaBuffer.get();
			}
			pinNumbers = new long[(int) pinCount];
			for (int i = 0; i < pinCount; i++) {
				int p = metaBuffer.getInt();
				// System.out.println("pin: "+p);
				pinNumbers[i] = uint32_tToLong(p);
			}
			printMetaData(oldStdOut);
			printMetaData(LogStdStreams.getLogStream());

			readAndConvertFiles(baseFileName);

			System.out.println(" --- Finished Processing");
			System.out.flush();
			LogStdStreams.getLogStream().println(" ---  Finished Processing");
			LogStdStreams.getLogStream().println();
			LogStdStreams.getLogStream().flush();

		} catch (Throwable t) {
			String msg = t.getMessage();
			if ((msg == null) || (msg.trim().length() == 0)) {
				msg = t.getClass().getSimpleName();
			}
			System.err.println(" Error! " + msg);
			t.printStackTrace(System.err);
		} finally {
			LogStdStreams.getLogStream().close();
		}
	}

	static void readAndConvertFiles(String baseName) {
		int fileNumber = 0;
		String fileNameBase = baseName;
		String fileName = "";
		for (fileNumber = 0; fileNumber <= 99; fileNumber++) { // final 99
			fileName = fileNameBase + (fileNumber < 10 ? "0" : "") + fileNumber;

			// System.out.println(fileName);
			// File binFile = new File(fileName);
			boolean resultOK = convertFileToCSV(fileName);
			if (!resultOK) {
				System.err.println(" Error! outputing CSV for "
						+ (new File(fileName + BIN_EXT)).getAbsolutePath());

			}
		}
	}

	static void putCsvHeader(PrintStream out) {
		out.print("SampleCounter (per " + sampleInterval + "uS)");
		for (int i = 0; i < pinCount; i++) {
			out.print(",");
			out.print("Pin " + pinNumbers[i]);
		}
		out.println();
	}

	// const size_t DATA_DIM16 = 254;
	// struct block16_t {
	// uint16_t count; // count of data values
	// uint16_t missedSamples; // count of missedSampless since last block
	// uint16_t data[DATA_DIM16];
	// };

	static boolean convertFileToCSV(String fileName) {
		boolean OK = true;
		// output progress msg

		String binFileName = fileName + BIN_EXT;
		String csvFileName = fileName + CSV_EXT;
		long missedSamplesThisFile = 0;
		FileOutputStream csvStream = null;
		PrintStream outStream = null;
		BufferedInputStream bufferedStream = null;
		FileInputStream binStream = null;
		File binFile = null;
		File csvFile = null;
		try {
			binFile = new File(binFileName);
			csvFile = new File(csvFileName);
			if (!binFile.exists()) {
				// skip this one
				return true;
			}
			if (csvFile.exists()) {
				System.err
						.println(" Error! Output .csv file already exists for ");
				System.err.println(" " + csvFile.getAbsolutePath());
				return false;
			}

			System.out.print("Processing " + binFile + "  ");
			LogStdStreams.getLogStream().print("Processing " + binFile + "  ");

			boolean skipCSVOutput = false; // set to true for testing only
			// to suppress csv sample output
			if (skipCSVOutput) {
				System.out.println(" skipping CSV output for testing..");
			}

			// read file in blocks of 512 bytes
			final int BUFFER_SIZE = 512;
			ByteBuffer binBuffer = ByteBuffer.allocate(BUFFER_SIZE);

			try {
				binStream = new FileInputStream(binFile);
			} catch (FileNotFoundException e) {
				System.err.println(" Error! Could not find data file");
				System.err.println(" " + binFile.getAbsolutePath());
				return false;
			}
			try {
				csvStream = new FileOutputStream(csvFile);
			} catch (Exception e) {
				System.err.println(" Error! Could not open output .csv file ");
				System.err.println(" " + csvFile.getAbsolutePath());
				System.err.println(" " + e.getMessage());
				OK = false;
			}
			if (csvStream != null) {

				// System.setOut(oldStdOut);
				// System.out.print("Processing " + fileName);
				// System.out.flush();
				// System.setOut(LogStdStreams.getLogStream()); // restore
				// redirected

				BufferedOutputStream bufferedCsvStream = new BufferedOutputStream(
						csvStream);
				outStream = new PrintStream(bufferedCsvStream);

				bufferedStream = new BufferedInputStream(binStream);

				putCsvHeader(outStream);

				int blockCounter = 0;
				int bytesRead = BUFFER_SIZE;
				while (bytesRead == BUFFER_SIZE) {
					try {
						bytesRead = bufferedStream.read(binBuffer.array(), 0,
								BUFFER_SIZE);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						System.err.println(" Error! Reading "
								+ binFile.getAbsolutePath());
						System.err.println(" " + e.getMessage());
						OK = false;
					}
					if (bytesRead != BUFFER_SIZE) {
						// did not read full buffer stop here
						break;
					}
					blockCounter++;
					if ((blockCounter % 4500) == 0) {
						System.out.print(".");
					}
					binBuffer.position(0);
					binBuffer.order(littleEndian ? ByteOrder.LITTLE_ENDIAN
							: ByteOrder.BIG_ENDIAN);
					int count = uint16_tToInt(binBuffer.getShort());
					int missedSamples = uint16_tToInt(binBuffer.getShort());
					// System.out.println("Block:" + blockCounter + " count:" +
					// count
					// + " missedSamples:" + missedSamples);
					if (missedSamples != 0) {
						System.out.println();
						System.out.print(" !!! Block:" + blockCounter
								+ " count:" + count + " missedSamples:"
								+ missedSamples);
						sampleIdx += missedSamples;
						missedSamplesThisFile += missedSamples;
					}
					// check that count is a whole number of samples
					int samples = (int) (count / pinCount);
					if (count % pinCount != 0) {
						System.err
								.println(" Error! Whole number of samples not stored. Processing "
										+ samples + " samples.");
						OK = false;
					}
					for (int i = 0; i < samples; i++) {
						if (!skipCSVOutput) {
							outStream.print(sampleIdx);
						}
						for (int k = 0; k < pinCount; k++) {
							if (!skipCSVOutput) {
								outStream.print(",");
							}
							int sample = uint16_tToInt(binBuffer.getShort());
							if (!skipCSVOutput) {
								outStream.print(sample);
							}
						}
						if (!skipCSVOutput) {
							outStream.println();
						}
						sampleIdx++; // time += sampleInterval;
					}

				}
				LogStdStreams.getLogStream().print(
						"  Processed " + blockCounter + " blocks");
				if (missedSamplesThisFile != 0) {
					LogStdStreams.getLogStream().println(
							" --- Total Missed Samples : "
									+ missedSamplesThisFile);
				}
				LogStdStreams.getLogStream().println();
				LogStdStreams.getLogStream().flush();

				//System.out.print("  Processed " + blockCounter + " blocks");
				if (missedSamplesThisFile != 0) {
					System.out.print(" Missed Samples : "
							+ missedSamplesThisFile);
				}
				System.out.println();
				System.out.flush();
				// System.setOut(LogStdStreams.getLogStream()); // restore
				// redirected

				if (bytesRead != -1) {
					System.err
							.println(" Error! - found partial block at block no:"
									+ blockCounter
									+ " in "
									+ binFile.getAbsolutePath());
					OK = false;
				}
			}
		} finally {
			if (bufferedStream != null) {
				try {
					bufferedStream.close();
				} catch (IOException e) {
					System.err.println(" Warning - error closing  "
							+ binFile.getAbsolutePath());
					System.err.println(" " + e.getMessage());
					OK = false;
				}
			}
			if (binStream != null) {
				try {
					binStream.close();
				} catch (IOException e) {
					System.err.println(" Warning - error closing  "
							+ binFile.getAbsolutePath());
					System.err.println(" " + e.getMessage());
					OK = false;
				}
			}
		}
		if (outStream.checkError()) {
			System.err.println(" Error! Error writing out .csv file "
					+ csvFile.getAbsolutePath());
			OK = false;
		}
		if (outStream != null) {
			outStream.flush();
		}

		if (csvStream != null) {
			try {
				outStream.close();
			} catch (Exception e) {
				System.err.println(" Warning - error closing  "
						+ binFile.getAbsolutePath());
				System.err.println(" " + e.getMessage());
				OK = false;
			}
		}

		return OK;
	}

	// prints time to parser
	static void mStoHrMinSec(long _mS, PrintStream out) {
	  long secs = (long)(_mS / 1000);
	  long mS = _mS - secs * 1000;
	  long mins = secs / 60;
	  secs = secs - mins * 60;
	  long hrs = mins / 60;
	  mins = mins - hrs * 60;
	  out.print(' ');
	  out.print(hrs); out.print("hrs ");
//	  if (mins <= 9) {
//	    out.print('0');
//	  }
	  out.print(mins); out.print("mins ");
//	  if (secs <= 9) {
//	    out.print('0');
//	  }
	  out.print(secs); out.print('.');
//	  if (mS <= 9) {
//	    out.print("00");
//	  } else if (mS <= 99) {
//	    out.print('0');
//	  }
	  out.print(mS);
	  out.print("s");
	}

	static void printMetaData(PrintStream out) {
		out.println(" Metadata Version : " + versionNo //
				+ "  (" + (littleEndian ? "Little Endian" : "Big Endian") + ")");

		// System.out.println("byteOrder:" + byteOrder); // 0x0000FFFF Note NOT
		// // Unicode marker which
		// // is 0000fffe
		out.println(" sampleInterval uS:" + sampleInterval); // Sample
																// interval
																// in
																// uS
		out.println(" adcBits:" + adcBits); // Size of ADC values.
		out.println(" adcAvgs:" + adcAvgs); // Number of ADC readings
											// averaged 0, 4, 8, 16 or
											// 32
		out.println(" pinCount:" + pinCount); // Number of analog pins in
												// a sample.
		out.print(" Pins: ");
		for (int i = 0; i < pinCount; i++) {
			out.print(pinNumbers[i]);
			if (i < (pinCount - 1)) {
				out.print(", ");
			}
		}
		out.println();
		out.println(" samplesPerBlock:" + samplesPerBlock); // =
															// SAMPLES_PER_BLOCK;
															// // i.e.
															// save
															// PIN_COUNT
															// *
		// SAMPLES_PER_BLOCK values in a block == block
		// count
		out.println(" noBufferBlocks:" + noBufferBlocks); // Number of
															// buffer
															// blocks
		out.print(" maxBuffersUsed:" + maxBuffersUsed); // max number
														// of
														// buffers
														// saved
														// at one
														// time
														// by
														// storeSampleBuffers()
		if (maxBuffersUsed > noBufferBlocks) {
			out.println("  Warning: Exceeds number of buffers available ("
					+ noBufferBlocks + "). Some samples may be missing.");
		} else {
			out.println();
		}
		out.print(" maxLatency uS:" + maxLatency); // max time spent in
													// storeSampleBuffers().
													// Includes
		if (maxLatency > noBufferBlocks * sampleInterval * samplesPerBlock) {
			out.println("  Warning: Exceeds time provided by buffer blocks ("
					+ (noBufferBlocks * sampleInterval * samplesPerBlock)
					+ "uS). Some samples will be missing.");
		} else {
			out.println();
		}
		// closing/opening files when swapping files uS
		out.println(" maxFileOpenTime uS:" + maxFileOpenTime); // max
																// time
																// spent
																// closing/opening
																// files
																// when
																// swapping
		// files uS
		out.println(" missedTimers:" + missedTimers);
		out.println(" missedSamplesTotal:" + missedSamplesTotal);
		out.println(" totalBlocksWritten:" + totalBlocksWritten);
		// out.println(" totalSamplesWritten:" + totalSamplesWritten);
		out.println(" totalSamplesWritten:" + totalSamplesWritten / pinCount);
		
		out.print(" requested runTime :");
		mStoHrMinSec(runTime,out); // how long to run this
		out.println();
		out.print(" run duration :");
		mStoHrMinSec(duration,out); // length of log data run, mS logging
		out.println();

		out.flush();
	}

	static long uint32_tToLong(int u) {
		long rtn = (long) u;
		rtn = rtn & 0x00000000ffffffffl; // clear leading 1's // NOTE the l, for
											// long, at the end of the hex
		return rtn;
	}

	static int uint16_tToInt(short s) {
		int rtn = (int) s;
		rtn = rtn & 0x0000ffff; // clear leading 1's
		return rtn;
	}

}
